/*
 * @(#)C_Demo.c    1.00   02/21/02
 *
 * Copyright (c) 2002-Present Agilis Software LLC, Santa Clara, CA, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Agilis Software LLC ("Confidential Information").  You shall not
 * disclose such Confidential Information and shall use it only
 * in accordance with the terms of the license agreement you entered
 * into with Agilis Software LLC.
 *
 * DISCLAIMER: this sample code is provided as-is.
 */

/*
 * This simple "hello EasyLicenser world" demo illustrates 
 * how to securely license-enable your application at a level one step above
 * the most basic level.  You can also use it as a command line utility for
 * analyzing a license key.
 *
 * Running the demo is equivalent to your end customer running your 
 * license-protected application at their site.  In order for the application to
 * run, you as an ISV will first need to generate a key at your premises with the 
 * EasyLicenser license manager GUI, publish it, and issue it to your
 * customer, who provides it to your application in a manner determined by you
 * and designed into your application.
 *
 * Features illustrated with this demo:
 * 1. How to use the "checkLicenseKey" API call.
 * 2. How to utilize the application-password security mechanism.
 * 3. How to protect against system clock changes without writing code.
 *
 * For examples on using custom key handlers for extensibility,
 * as well as using digital signatures and other anti-spoofing measures,
 * see the accompanying "cdemosecure.cpp" demo application.
 *
 * For examples on securely managing license state information such as 
 * usage quotas, see the accompanying "cdemostate.cpp" demo application.
 *
 * Some important points to note about this demo:
 * 1. The code that is directly relevant to using the EasyLicenser API's
 *    is indicated as such in comments.  The rest of the code exists in
 *    order to have a functioning self-contained program.
 * 2. All code pertaining to reading the key from an EasyLicenser file is
 *    for functional purposes only.  It is not meant to suggest a specific
 *    technique for managing the key in your application.  Nor is it
 *    a part of the EasyLicenser product.  In particular, note that your
 *    actual application will typically have access to just the license key
 *    without any accompanying metadata about the key.
 * 3. The demo will work with keys generated with EasyLicenser 2.0 as well
 *    as prior versions.  The EasyLicenser 2.0 run time library is required.
 * 4. Although the demo performs an introspection on the license key context
 *    following a successful license key check, the degree of introspection
 *    that your actual application will perform will depend on your
 *    specific requirements and will usually range from no introspection to 
 *    introspection on just the product options.
 * 5. This demo does not perform automatic state management like "cdemostate.cpp"
 *    does.  Therefore, even if the key you supply is quota limited, you will
 *    not get an "out of quota" error condition regardless of the number of
 *    times you run the demo.
 */

#define __STRICT_ANSI__ 1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>

#include <ezLicenseInfo.h>

/* File types */
#define EZLM_FILE_TYPE_UNKNOWN	0
#define EZLM_FILE_TYPE_TXT		1
#define EZLM_FILE_TYPE_XML		2
#define EZLM_FILE_TYPE_CSV		3
#define EZLM_FILE_TYPE_TAB		4
#define EZLM_FILE_TYPE_INI		5
#define EZLM_FILE_TYPE_KEY		6

void CALLING_CONVENTION displayWarnings(int iWarningsBitMap);
void CALLING_CONVENTION displayError(int errorCode);
void CALLING_CONVENTION displayLicenseParams(void* ezLicInstance);
int CALLING_CONVENTION getFileType(const char *fname);
int CALLING_CONVENTION getKey(const char *filelines[], int numlines,
							  int currLine, const int fileType,
							  char keybuf[], size_t keybufsize);
int CALLING_CONVENTION readFile(const char *filename, char *filelines[], int maxfilelines);
int CALLING_CONVENTION strip(const char *instr, char retbuf[], size_t retbufsize);

/**
 * main entry point:
 * Accepts two mandatory command line arguments:
 * 1. Name of a text file containing a license key to be validated.
 * 2. Expected user / host name that should match what's in the key.
 *    If the key is known to have been generated with the "enforce OS
 *    user / host" flag, you can specify the special string "null" to
 *    signify a null value.
 * and two optional arguments (either none or both must be specified):
 * 3. Case-sensitive product name that was specified at the time of key creation,
 *    if the product had an associated application password
 * 4. Product's associated password's case-sensitive public key, available with
 *    the product definition and also obtainable with the "makepubkey" Java utility
 *    located at ezlic2.0/util.
 */
int main(int argc, char **argv)
{
	FILE* fptr = 0;
	char 	*strLicenseKeyFileName = NULL;
	char 	*strLicenseKey = NULL;
	int		iDaysLeftThreshold = 7;
	long	lQuotaUsageToDate = 0;
	long	lQuotaThreshold = 10;
	int		iDaysGracePeriod = 365; 
	long	lQuotaGrace = 100;
	long	lCurrUsageValue = 0;
	int		iStatusCode = 0;
	char	*strUserHostName;
	void *ezLicInfoInstance = 0;
	char	*suffix = NULL;
	char *filelines[200];
	int numlines = 0;
	char keybuf[MAX_ENCRYPTED_LICENSE_TEXT_LENGTH];
	char *appPwd = NULL;
	char *productName = NULL;
	int fileType = EZLM_FILE_TYPE_UNKNOWN;
	int currLine = 0;
	int keyno = 0;

    char    ezlmMajorVersion[MAX_VERSION_INFO_LENGTH];
    char    ezlmMinorVersion[MAX_VERSION_INFO_LENGTH];

	if (argc < 3 || argc == 4) {
		printf("\n*** usage: %s key-text-file-name user-or-host-name [product-name application-password-public-key] ***", argv[0]);
		printf("\n    Example: %s keypub.csv user1\n", argv[0]);
		exit(2);
	}

	strLicenseKeyFileName = argv[1];
	strUserHostName = argv[2];
	if (argc > 3) {
		productName = argv[3];
		appPwd = argv[4];
	}

	printf("\nHello EasyLicenser World.\n");
    printf("\nRuntime library version information:\n");
    if (getEzlmVersionInfoStr(ezlmMajorVersion, sizeof(ezlmMajorVersion),
                              ezlmMinorVersion, sizeof(ezlmMinorVersion)) < 0) {
        printf("*** Error getting EasyLicenser runtime version information ***\n");
        exit (2);
    }
    printf("  Major Version: %s, Minor Version: %s\n", ezlmMajorVersion, ezlmMinorVersion);

	printf("\nCommand line arguments:\n");
	printf("  License key file:          %s\n", strLicenseKeyFileName);
	printf("  User / Host name:          %s\n", strUserHostName);
	printf("  Product name:              %s\n", (productName == NULL ? "<not specified>":productName));
	printf("  App Passwd Public Key:     %s\n", (appPwd == NULL ? "<not specified>":appPwd));

	if (strcmp(strUserHostName,"null")==0)
		strUserHostName = NULL;

	fileType = getFileType(strLicenseKeyFileName);

	numlines = 
		readFile(strLicenseKeyFileName, filelines, 1000);
	if (numlines < 0) {
		printf("*** Cannot process license key file due to errors.  Aborting. ***\n");
		exit(1);
	} else if (!numlines) {
		printf("License key file is empty.  Nothing to do.  Exiting.\n");
		exit(0);
	}

	printf("\n========================================================\nChecking all license key in file...\n");
	strLicenseKey = &keybuf[0];
	currLine = 0;
	keyno = 0;
	while (currLine < numlines &&
	       (currLine = getKey((const char **)filelines, numlines, currLine, fileType, keybuf, sizeof(keybuf))) > 0) {
		keyno++;
		printf("\nLicense key #%d:\n----------------\n", keyno);
		printf("[%s]\n\n", keybuf);

		/* EasyLicenser API call (required): allocate a license context. */
		ezLicInfoInstance = allocLicenseInfo();
		/* EasyLicenser API call (required): perform the license key check. */
		iStatusCode = checkLicenseKey_p(
						ezLicInfoInstance,	/* license context */
						keybuf, 			/* the license key string */
						0, 0,				/* no custom key handler  */
											/* for this simple example */
						iDaysLeftThreshold, /* threshold days to trigger  */
											/* "about to expire" warning */
						lQuotaUsageToDate,	/* quota consumption to date */
						lQuotaThreshold, 	/* threshold balance quota to */
											/* trigger "about to exhaust */
											/* quota" warning */
						iDaysGracePeriod,	/* allow license key check to */
											/* succeed with warning instead */
											/* of error within these many */
											/* days past expiration date. */
						lQuotaGrace,		/* allow license key check to */
											/* succeed with warning instead */
											/* of error within these many */
											/* uses past quota limit. */
						strUserHostName, 	/* user / host name - can be null */
											/* if OS enforcement is in effect. */
						lCurrUsageValue,	/* current use level (note: different */
											/* from quota which is cumulative */
											/* and can use different units) - */
											/* checked against limit in key. */
											/* applies only to non-single-user */
											/* licenses. */
						productName,		/* name of product for which key */
											/* was generated */
						appPwd);			/* application password public key, */
											/* if the product had an application */
											/* password specified when the key was */
											/* generated. */

		if (iStatusCode < 0) {
			/* license check failed: report the error condition */
			displayError(iStatusCode);
		} else {	
			/* license check succeeded.  If the return value is > 0, */
			/* it means there were warnings issued - report these */
			printf("    License key validation successful%s.\n\n", (iStatusCode?" with warnings":""));
			/* introspect the entire license key contents - */
			/* not something you'd do in an actual application */
			displayLicenseParams(ezLicInfoInstance);
			displayWarnings(iStatusCode);
		}
		/* EasyLicenser API call (required): release the license context. */
		freeLicenseInfo(ezLicInfoInstance);
	}
	printf("\n========================================================\n\nAll keys checked.\n\n");
	return 0;
}


/*
*	Display warnings if any.  Warning codes are defined in <ezLicStatusCodes.h>
*	@param iWarningsBitMap : Warning bitMap returned from license key API check.
*/
void CALLING_CONVENTION displayWarnings(int iWarningsBitMap)
{
	if (!iWarningsBitMap)
		return;
	printf("        Warning codes:\n");
    if (iWarningsBitMap & EZLIC_WARN_EXPIRING)
        printf("          Time limited license key is about to expire.\n");
	if (iWarningsBitMap & EZLIC_WARN_QUOTA)
        printf("          Quota limited license key is about to expire.\n");
	if (iWarningsBitMap & EZLIC_WARN_EXPIRED)
        printf("          Time limited license has expired and is now in grace period.\n");
	if (iWarningsBitMap & EZLIC_WARN_QUOTA_EXCEEDED)
        printf("          Quota limited license key has expired and is now in grace period.\n");
	if (iWarningsBitMap & EZLIC_WARN_EXP_GRACE_TRUNC)
		printf("          Expiration Days Grace In API Call Too Large - Truncated To 5 / 25 pct Of Lifespan\n");
	if (iWarningsBitMap & EZLIC_WARN_QUOTA_GRACE_TRUNC)
		printf("          Quota Grace In API Call Too Large - Truncated To 5 / 25 pct Of Original Quota\n");
}


/*
*	Display the meaning of the -ve error code returned from the
*   license key API check.  Error codes are defined in <ezLicStatusCodes.h>
*/
void CALLING_CONVENTION displayError(int errorCode)
{
	if (!errorCode)
		return;

	printf("*** ");
	switch(errorCode) {
	case     EZLIC_EXC_BASE:
			printf("Error: General license validation failure. Please check your license key.\n");
			break;
	case     EZLIC_EXC_MALFORMED_KEY:
			printf("Error: the license key is invalid.\n");
			break;
	case     EZLIC_EXC_EXPIRED:
			printf("Error: The license has expired past the grace period.\n");
			break;
	case     EZLIC_EXC_EXCEEDED_LIMITS:
			printf("Error: The licensed user / server / CPU count / quota limit in the key has been exceeded.\n");
			break;
	case     EZLIC_EXC_INVALID_HOST_USER_NET:
			printf("Error: The user / machine name in the license key does not match your environment / parameter.\n");
			break;
	case     EZLIC_EXC_INVALID_CUSTOM_KEY:
			printf("Error: The custom key contained in the license key is invalid.\n");
			break;
	case     EZLIC_EXC_UNIMPLEMENTED:
			printf("Error: The feature you are using is not implemented.\n");
			break;
	case     EZLIC_EXC_GEN:
			printf("Error: Key generation execption (you shouldn't encounter this condition).\n");
			break;
	case     EZLIC_EXC_VENDOR:
			printf("Error: Vendor license key exception (you shouldn't encounter this condition).\n");
			break;
	case     EZLIC_EXC_INTERNAL:
			printf("Error: internal error in EasyLicenser - please submit a bug report.\n");
			break;
	case     EZLIC_EXC_OS:
			printf("Error: An operating system error occurred - please check your machine configuration.\n");
			break;
	case     EZLIC_EXC_CONFIG:
			printf("Error: exception in system configuration - please check your EasyLicenser installation and system clock.\n");
			break;
	case     EZLIC_EXC_UNLICENSED_OPTION:
			printf("Error: you attempted to perform an unlicensed operation.\n");
			break;
	case     EZLIC_EXC_USAGE:
			printf("Error: API usage - please check your API call parameters and call sequence.\n");
			break;
	case     EZLIC_EXC_AUTHENT:
			printf("Error: Key / cookie authentication failure - please check your product name and application password.\n");
			break;
	default:
			printf("Error: unknown error code %d.\n", errorCode);
			break;
	}
}

void CALLING_CONVENTION showLicenseModel(int licenseModel,
										 int licenseType)
{
	printf("        License Model: %s\n", 
			(licenseModel == EZLIC_MODEL_USER ? "User":
			 licenseModel == EZLIC_MODEL_SERVER ? "Server":
			 "*** Unknown ***"));
	printf("        License Type: ");
	switch(licenseModel) {
	case EZLIC_MODEL_USER:
		switch(licenseType) {
		case EZLIC_TYPE_U_USER:
			printf("Named User\n");
			break;
		case EZLIC_TYPE_U_NODE:
			printf("Node Locked\n");
			break;
		case EZLIC_TYPE_U_FLOAT:
			printf("Floating User\n");
			break;
		default:
			printf("*** Unknown license type %d ***\n", licenseType);
		}
		break;
	case EZLIC_MODEL_SERVER:
		switch(licenseType) {
		case EZLIC_TYPE_SVR_CONC:
			printf("Concurrent User\n");
			break;
		case EZLIC_TYPE_SVR_CPU:
			printf("CPU Count\n");
			break;
		case EZLIC_TYPE_SVR_MHZ:
			printf("CPU MHZ\n");
			break;
		case EZLIC_TYPE_SVR_NMU:
			printf("Named User\n");
			break;
		default:
			printf("*** Unknown license type %d ***\n", licenseType);
		}
		break;
	default:
		printf("%d\n", licenseModel);
	}
}


/*
* Introspect the license key and display its parameters.
* In an actual application you may introspect specific
* values and act on them instead of displaying them.
* The most common parameter that is post-processed is
* the product options.
*/
void CALLING_CONVENTION displayLicenseParams(
							void* ezLicInstance)
{
	int value = 0;
	long lvalue = 0L;
	long ltime = 0;
	char* strValue = 0;


	printf("    License Key Information:\n");
	printf("        Key version:    ");
	printf((getKeyVersion(ezLicInstance) <= 0 ? "Legacy (1.1 or earlier)\n":"%d\n"),
		   getKeyVersion(ezLicInstance));
	strValue = getChainedKeyId(ezLicInstance);
	if (strValue && strValue[0])
		printf("        Replacement for key with Id: '%s'\n", strValue);

	value = getLicenseModelCode(ezLicInstance);
	showLicenseModel(value, getLicenseTypeCode(ezLicInstance));
	if (getEnforce(ezLicInstance))
		printf("        Operating System check enforced.\n");

	value = getLicenseModeBitmap(ezLicInstance);
	printf("        License Mode Bitmap:    0x%x\n", value);
	if (value & EZLIC_MODE_TIME)
		printf("          TIME LIMITED\n");
	if (value & EZLIC_MODE_OPTIONS)
		printf("          OPTIONS\n");
	if (value & EZLIC_MODE_METERED)
		printf("          METERED\n");
	if (value & EZLIC_MODE_CUSTOM_KEY)
		printf("          CUSTOM KEY\n");
	if (value & EZLIC_MODE_CUSTOM_COOKIE)
		printf("          CUSTOM COOKIE\n");

	lvalue = getUsageValue(ezLicInstance);
	if (lvalue > 0)
		printf("        Usage Limit Value:    %ld\n", lvalue);

	lvalue = getQuotaValue(ezLicInstance);
	if (lvalue > 0)
		printf("        Quota Limited To:    %ld\n", lvalue);

	strValue = getOptions(ezLicInstance);
	if (strValue  && strValue[0])
		printf("        Options:    [%s]\n", strValue);

	if ((ltime = getExpirationTime(ezLicInstance)) > 0) {
		char tmpbuf[100];
		struct tm when;
		when = *gmtime( &ltime );
		strcpy(tmpbuf,asctime(&when));
		if (tmpbuf[strlen(tmpbuf)-1]=='\n')
			tmpbuf[strlen(tmpbuf)-1] = '\0';
		printf( "        Time Limited To Expire (GMT):    %s\n", 
				tmpbuf);
	}
   	
	strValue = getUserHostNetName(ezLicInstance);
	if (strValue && strValue[0])
		printf("        User/Host Name:    [%s]\n", strValue);
	strValue = getCustomCookie(ezLicInstance);
	if (strValue && strValue[0])
		printf("        Custom Cookie:    [%s]\n", strValue);
	strValue = getCustomKey(ezLicInstance);
	if (strValue && strValue[0])
		printf("        Custom Key:    [%s]\n", strValue);
	strValue = getAppPwd(ezLicInstance);
	if (strValue && strValue[0])
		printf("        Application Password:    [%s]\n", strValue);
}


/* 
 * The file IO is for convenience purposes only - you'd manage the persistence of 
 * your key according to the requirements of your application.
 */

int CALLING_CONVENTION readFile(const char *filename, char *filelines[], int maxfilelines)
{
	int numlines = 0;
	int totlen = 0;
	int len = 0;
	char linebuf[10240];
	char *linep = NULL;
	char *lp = NULL;
	int trunc = 0;
	FILE *fh = fopen(filename,"r");
	if (!fh) {
		printf("\n*** Cannot open license key file '%s' ***\n",
				filename);
		return -1;
	}
	while (numlines <= maxfilelines &&
	       fgets(linebuf, sizeof(linebuf), fh)) {
		linep = (char *)malloc(sizeof(linebuf));
		lp = &linebuf[0];
		while (*lp && isspace((int)*lp))
			lp++;
		if (*lp) {
			strcpy(linep, lp);
			filelines[numlines++] = linep;
		}
	}
	if ((!trunc) && numlines > maxfilelines)
		printf("\n*** Warning: IO buffer overflow, truncating.\n");
	fclose(fh);
	return numlines;
}


int CALLING_CONVENTION getKey(const char *filelines[], int numlines,
							  int currLine, const int fileType,
							  char keybuf[], size_t keybufsize)
{
	const char *keyp = NULL;

	if (currLine >= numlines)
		return 0;

	if (currLine == 0) {
		/* check / adjust for first line */
		switch (fileType) {
		case EZLM_FILE_TYPE_TXT:
			if (strncmp(filelines[currLine++], "License Key List:", strlen("License Key List:")) ||
			    strncmp(filelines[currLine++], "----------------", strlen("----------------"))) {
				printf("*** Not a valid EasyLicenser text export file ***\n");
				exit(2);
			}
			currLine += 3; /* position at "Key Id:" */
			break;
		case EZLM_FILE_TYPE_XML:
			if (strncmp(filelines[currLine++], "<?xml", strlen("<?xml")) ||
			    strncmp(filelines[currLine++], "<key-list>", strlen("<key-list>"))) {
				printf("*** Not a valid EasyLicenser XML export file ***\n");
				exit(2);
			}
		default:
			break;
		}
	} else {
		/* check for EOF */
		switch (fileType) {
		case EZLM_FILE_TYPE_XML:
			if (filelines[currLine][0] && strncmp(filelines[currLine],"</key-list>",strlen("</key-list>"))==0)
				return 0;
		default:
			if (!filelines[currLine][0])
				return 0;
			break;
		}
	}
	/* get key */
	switch (fileType) {
	case EZLM_FILE_TYPE_TXT:
		for (; strncmp(filelines[currLine], "License key:", strlen("License key:")) && 
		       currLine < numlines; 
		     currLine++);
		if (++currLine >= numlines) {
			printf("*** Premature termination of text file ***\n");
			return 0;
		}
		keyp = filelines[currLine];
		strip(keyp, keybuf, keybufsize);
		currLine += 2;
		break;
	case EZLM_FILE_TYPE_XML:
		if (strncmp(filelines[currLine], "<key id=", strlen("<key id="))) {
			printf("*** Invalid XML record - treating as EOF ***\n");
			return 0;
		}
		if (++currLine >= numlines) {
			printf("*** Premature termination of XML file ***\n");
			return 0;
		}
		keyp = filelines[currLine];
		strip(keyp, keybuf, keybufsize);
		currLine++;
		currLine++;
		break;
	case EZLM_FILE_TYPE_INI:
		if (!(keyp = strchr(filelines[currLine], (int)'='))) {
			printf("*** Invalid property file format - treating as EOF ***\n");
			return 0;
		}
		strip(keyp+1, keybuf, keybufsize);
		currLine++;
		break;
	case EZLM_FILE_TYPE_KEY:
		strip(filelines[currLine], keybuf, keybufsize);
		currLine++;
		break;
	case EZLM_FILE_TYPE_CSV:
	case EZLM_FILE_TYPE_TAB:
		keyp = strrchr(filelines[currLine], (fileType==EZLM_FILE_TYPE_CSV?(int)',':(int)'\t'));
		if (!keyp) {
			printf("*** Invalid %s file format - treating as EOF ***\n",
					(fileType==EZLM_FILE_TYPE_CSV?"csv":"tab"));
			return 0;
		}
		strip(keyp+1, keybuf, keybufsize);
		currLine++;
		break;
	}
	return currLine;
}


int CALLING_CONVENTION getFileType(const char *fname) {
	char *suffix = strrchr(fname, (int)'.');
	if (suffix > fname && suffix > &fname[strlen(fname)-4])
		return EZLM_FILE_TYPE_UNKNOWN;
	suffix++;
	if (strcmp(suffix, "txt")==0)
		return EZLM_FILE_TYPE_TXT;
	if (strcmp(suffix, "xml")==0)
		return EZLM_FILE_TYPE_XML;
	if (strcmp(suffix, "csv")==0)
		return EZLM_FILE_TYPE_CSV;
	if (strcmp(suffix, "tab")==0)
		return EZLM_FILE_TYPE_TAB;
	if (strcmp(suffix, "ini")==0)
		return EZLM_FILE_TYPE_INI;
	if (strcmp(suffix, "key")==0)
		return EZLM_FILE_TYPE_KEY;
	printf("*** Unknown file type: '%s'.  Exiting ***\n", suffix);
	exit(1);
}


/* 
 * Strip the key of all leading, trailing and embedded white spaces.
 * A good idea for an actual application, too.
 */
int CALLING_CONVENTION strip(const char *instr, char retbuf[], size_t retbufsize)
{
	char *outptr = &retbuf[0];
	int len = 0;
	char ch = 0;
	if (retbufsize <= strlen(instr))
		return -1;
	while (ch = *instr++) {
		if (!isspace((int)ch))
			*outptr++ = ch;
	}
	*outptr = '\0';
	return 0;
}

